import torch.nn as nn
import sys
from ..blocks.cnn_blocks import ConvBlock, ResBlock, Concat, Sum


class CGP2CNN(nn.Module):
    def __init__(self, cgp, in_channel, n_class, imgSize):
        super(CGP2CNN, self).__init__()
        self.cgp = cgp
        self.channel_num = [None for _ in range(500)]
        self.size = [None for _ in range(500)]
        self.channel_num[0] = in_channel
        self.size[0] = imgSize
        self.encode = nn.ModuleList()
        self.n_class = n_class
        self._build_network()

    def _build_network(self):
        for i, gene in enumerate(self.cgp):
            if len(gene) == 3:
                name, in1, in2 = gene
            elif len(gene) == 2:
                name, in1 = gene
                in2 = None
            else:
                raise ValueError(f"Unexpected number of elements in gene: {gene}")

            if name == 'input':
                continue
            elif name == 'full':
                self._add_full_layer(i, in1)
            elif name in ['Max_Pool', 'Avg_Pool']:
                self._add_pooling_layer(i, in1, name)
            elif name == 'Concat':
                self._add_concat_layer(i, in1, in2)
            elif name == 'Sum':
                self._add_sum_layer(i, in1, in2)
            else:
                self._add_block_layer(i, name, in1)

    def _add_full_layer(self, i, in1):
        self.encode.append(nn.Linear(self.channel_num[in1] * self.size[in1] * self.size[in1], self.n_class))

    def _add_pooling_layer(self, i, in1, name):
        self.channel_num[i] = self.channel_num[in1]
        self.size[i] = self.size[in1] // 2
        self.encode.append(nn.MaxPool2d(2, 2) if name == 'Max_Pool' else nn.AvgPool2d(2, 2))

    def _add_concat_layer(self, i, in1, in2):
        if in2 is None:
            raise ValueError(f"Concat operation requires two input indices, got: {in1}")
        self.channel_num[i] = self.channel_num[in1] + self.channel_num[in2]
        self.size[i] = self.size[min(in1, in2, key=lambda x: self.size[x])]
        self.encode.append(Concat())

    def _add_sum_layer(self, i, in1, in2):
        if in2 is None:
            raise ValueError(f"Sum operation requires two input indices, got: {in1}")
        self.channel_num[i] = max(self.channel_num[in1], self.channel_num[in2])
        self.size[i] = self.size[min(in1, in2, key=lambda x: self.size[x])]
        self.encode.append(Sum())

    def _add_block_layer(self, i, name, in1):
        try:
            down, func, out_size, kernel = name.split('_')
            out_size, kernel = int(out_size), int(kernel)
        except ValueError:
            raise ValueError(f"Invalid format for operation name: {name}")

        if down == 'S':
            if func == 'ConvBlock':
                self.channel_num[i] = out_size
                self.size[i] = self.size[in1]
                self.encode.append(ConvBlock(self.channel_num[in1], out_size, kernel, stride=1))
            elif func == 'ResBlock':
                self.channel_num[i] = max(out_size, self.channel_num[in1])
                self.size[i] = self.size[in1]
                self.encode.append(ResBlock(self.channel_num[in1], out_size, kernel, stride=1))
            else:
                raise ValueError(f"Unsupported block type: {func}")
        else:
            raise ValueError(f"Unsupported operation: {down}")

    def forward(self, x):
        outputs = [None for _ in range(len(self.cgp))]
        outputs[0] = x
        for i, layer in enumerate(self.encode):
            if isinstance(layer, ConvBlock):
                outputs[i + 1] = layer(outputs[self.cgp[i + 1][1]])
            elif isinstance(layer, ResBlock):
                outputs[i + 1] = layer(outputs[self.cgp[i + 1][1]], outputs[self.cgp[i + 1][1]])
            elif isinstance(layer, nn.Linear):
                tmp = outputs[self.cgp[i + 1][1]].view(outputs[self.cgp[i + 1][1]].size(0), -1)
                outputs[i + 1] = layer(tmp)
            elif isinstance(layer, (nn.MaxPool2d, nn.AvgPool2d)):
                if outputs[self.cgp[i + 1][1]].size(2) > 1:
                    outputs[i + 1] = layer(outputs[self.cgp[i + 1][1]])
                else:
                    outputs[i + 1] = outputs[self.cgp[i + 1][1]]
            elif isinstance(layer, Concat):
                outputs[i + 1] = layer(outputs[self.cgp[i + 1][1]], outputs[self.cgp[i + 1][2]])
            elif isinstance(layer, Sum):
                outputs[i + 1] = layer(outputs[self.cgp[i + 1][1]], outputs[self.cgp[i + 1][2]])
            else:
                sys.exit("Error in forward pass of CGP2CNN")
        return outputs[-1]



import time
import numpy as np
import math
import matplotlib.pyplot as plt
from ..cgp.individual import Individual
import os


class CGP:
    def __init__(self, net_info, eval_func, lam, imgSize, verbose, logger, result_folder_path):
        self.lam = lam
        self.pop = [Individual(net_info, False) for _ in range(1 + self.lam)]
        self.eval_func = eval_func
        self.num_gen = 0
        self.num_eval = 0
        self.max_pool_num = int(math.log2(imgSize) - 2)
        self.logger = logger
        self.best_accuracy = -float('inf')
        self.verbose = verbose
        self.result_folder_path = result_folder_path
        self.history = []

        os.makedirs(self.result_folder_path, exist_ok=True)

    def _evaluation(self, pop, eval_flag):
        net_lists = [pop[i].active_net_list() for i in np.where(eval_flag)[0]]
        fp = self.eval_func(net_lists)
        for i, j in enumerate(np.where(eval_flag)[0]):
            pop[j].eval = fp[i]
        return np.array([ind.eval for ind in pop])

    def modified_evolution(self, max_eval, mutation_rate):
        start_time = time.time()
        eval_flag = np.empty(self.lam)
        active_num = self.pop[0].count_active_node()
        _, pool_num = self.pop[0].check_pool()

        while active_num < self.pop[0].net_info.min_active_num or pool_num > self.max_pool_num:
            self.pop[0].mutation(1.0)
            active_num = self.pop[0].count_active_node()
            _, pool_num = self.pop[0].check_pool()

        evaluations = self._evaluation([self.pop[0]], np.array([True]))
        self.history.append(evaluations[evaluations.argmax()])

        while self.num_gen < max_eval:
            self.num_gen += 1
            for i in range(self.lam):
                eval_flag[i] = False
                self.pop[i + 1].copy(self.pop[0])
                active_num = self.pop[i + 1].count_active_node()
                _, pool_num = self.pop[i + 1].check_pool()
                while not eval_flag[i] or active_num < self.pop[i + 1].net_info.min_active_num or pool_num > self.max_pool_num:
                    self.pop[i + 1].copy(self.pop[0])
                    eval_flag[i] = self.pop[i + 1].mutation(mutation_rate)
                    active_num = self.pop[i + 1].count_active_node()
                    _, pool_num = self.pop[i + 1].check_pool()

            evaluations = self._evaluation(self.pop[1:], eval_flag=eval_flag)
            best_arg = evaluations.argmax()
            best_accuracy = evaluations[best_arg]

            if best_accuracy > self.best_accuracy:
                self.best_accuracy = best_accuracy
                self.best_model = self.pop[best_arg + 1]

            if evaluations[best_arg] > self.pop[0].eval:
                self.pop[0].copy(self.pop[best_arg + 1])
            else:
                self.pop[0].neutral_mutation(mutation_rate)

            self.history.append(self.best_accuracy)

            if self.verbose:
                self.logger.info(f'Generation {self.num_gen}: Best Model Accuracy = {best_accuracy:.4f} History: {self.history}')
                plt.plot(self.history)
                plt.xlabel('Generations')
                plt.ylabel('Best Fitness')
                plt.savefig(os.path.join(self.result_folder_path, f'generation_fitness_plot.png'))

        if self.verbose:
            total_time = time.time() - start_time
            self.logger.info(f'Total Evolution Time: {total_time:.2f} seconds')

        return self.best_model, self.best_accuracy, self.history



# src/loaders/image_net_scraper.py

import os
import requests
import json
import torch
import shutil
from torchvision import transforms
from multiprocessing import Pool, cpu_count
from requests.exceptions import ConnectionError, ReadTimeout, TooManyRedirects, MissingSchema, InvalidURL
import pandas as pd
from .singleton import Singleton


class ImageNetScraper(metaclass=Singleton):
    def __init__(self, input_filepath, split_proportions, image_size, class_list, images_per_class, multiprocessing_workers, logger):
        self.images_per_class = images_per_class
        self.class_list = class_list
        self.multiprocessing_workers = multiprocessing_workers if multiprocessing_workers else cpu_count()
        self.logger = logger

        self.class_info_dict = self.load_classes_from_json()
        self.class_info_df = self.load_classes_from_csv()

        self.input_filepath = input_filepath
        self.split_proportions = split_proportions
        self.image_size = image_size
        self.class_names = self.class_info_df[self.class_info_df['synid'].isin(self.class_list)].set_index('synid')['class_name'].to_dict()

        self.transform = transforms.Compose([
            transforms.Resize(self.image_size),
            transforms.ToTensor()
        ])

    def load_classes_from_json(self):
        json_filepath = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', 'resources', 'imagenet_class_info.json')
        with open(json_filepath) as file:
            return json.load(file)

    def load_classes_from_csv(self):
        csv_filepath = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', 'resources', 'classes_in_imagenet.csv')
        return pd.read_csv(csv_filepath)

    def get_image(self, img_url, class_name, class_images_counter):
        response = requests.get(img_url, timeout=1)
        if 'image' not in response.headers.get('content-type', ''):
            self.logger.error("Not an image")
            raise ValueError("Not an image")
        img_content = response.content
        if len(img_content) < 1000:
            self.logger.error("Image too small")
            raise ValueError("Image too small")
        img_file_path = os.path.join(self.input_filepath, f'{class_name}_{class_images_counter}.png')
        with open(img_file_path, 'wb') as img_f:
            img_f.write(img_content)
        self.logger.info(f"Saved image {img_file_path}")

    def download_images(self, urls, class_name):
        class_images_counter = 0
        for url in urls:
            if class_images_counter >= self.images_per_class:
                self.logger.info(f"Reached the limit of {self.images_per_class} images for class {class_name}.")
                break
            try:
                class_images_counter += 1
                self.get_image(url, class_name, class_images_counter)
            except (ConnectionError, ReadTimeout, TooManyRedirects, MissingSchema, InvalidURL, ValueError) as e:
                self.logger.error(f"Failed to download image: {e}")
                class_images_counter -= 1
        self.logger.info(f"Downloaded images for class {class_name}.")

    def fetch_image_urls(self, wnid):
        url = f'http://www.image-net.org/api/imagenet.synset.geturls?wnid={wnid}'
        response = requests.get(url)
        return [url.decode('utf-8') for url in response.content.splitlines()]

    def scrape_class(self, class_wnid):
        class_name = self.class_info_dict[class_wnid]["class_name"]
        img_urls = self.fetch_image_urls(class_wnid)
        self.logger.info(f'Starting download for class "{class_name}" with a limit of {self.images_per_class} images.')
        self.download_images(img_urls, class_name)

    def run(self):
        with Pool(self.multiprocessing_workers) as pool:
            pool.map(self.scrape_class, self.class_list)

    def load_and_split_data(self):
        self.logger.info("Starting to run scraper to download images.")

        self.train_path = os.path.join(self.input_filepath, 'train')
        self.valid_path = os.path.join(self.input_filepath, 'valid')
        self.test_path = os.path.join(self.input_filepath, 'test')

        os.makedirs(self.train_path, exist_ok=True)
        os.makedirs(self.valid_path, exist_ok=True)
        os.makedirs(self.test_path, exist_ok=True)

        self.logger.info("Created directories")

        self.run()

        images = [os.path.join(self.input_filepath, file) for file in os.listdir(self.input_filepath) if file.endswith(('.png', '.jpg', '.jpeg'))]
        total_images = len(images)
        self.logger.info(f"Found {total_images} images for processing.")

        train_size = int(total_images * self.split_proportions[0])
        valid_size = int(total_images * self.split_proportions[1])

        torch.manual_seed(0)
        shuffled_indices = torch.randperm(total_images)

        for class_name in self.class_names.values():
            os.makedirs(os.path.join(self.train_path, class_name), exist_ok=True)
            os.makedirs(os.path.join(self.valid_path, class_name), exist_ok=True)
            os.makedirs(os.path.join(self.test_path, class_name), exist_ok=True)

        for idx in shuffled_indices[:train_size]:
            image_path = images[idx]
            class_name = os.path.basename(image_path).split('_')[0]
            dest_path = os.path.join(self.train_path, class_name, os.path.basename(image_path))
            shutil.move(image_path, dest_path)
            self.logger.info(f"Moved {image_path} to {dest_path}")

        for idx in shuffled_indices[train_size:train_size + valid_size]:
            image_path = images[idx]
            class_name = os.path.basename(image_path).split('_')[0]
            dest_path = os.path.join(self.valid_path, class_name, os.path.basename(image_path))
            shutil.move(image_path, dest_path)
            self.logger.info(f"Moved {image_path} to {dest_path}")

        for idx in shuffled_indices[train_size + valid_size:]:
            image_path = images[idx]
            class_name = os.path.basename(image_path).split('_')[0]
            dest_path = os.path.join(self.test_path, class_name, os.path.basename(image_path))
            shutil.move(image_path, dest_path)
            self.logger.info(f"Moved {image_path} to {dest_path}")

        for class_name in self.class_names.values():
            train_class_path = os.path.join(self.train_path, class_name)
            valid_class_path = os.path.join(self.valid_path, class_name)
            test_class_path = os.path.join(self.test_path, class_name)

            if os.path.exists(train_class_path) and len(os.listdir(train_class_path)) == 0:
                os.rmdir(train_class_path)

            if os.path.exists(valid_class_path) and len(os.listdir(valid_class_path)) == 0:
                os.rmdir(valid_class_path)
                self.logger.info(f"Removed empty directory {valid_class_path}")

            if os.path.exists(test_class_path) and len(os.listdir(test_class_path)) == 0:
                os.rmdir(test_class_path)

    def clear_directories(self):
        directory = self.input_filepath
        if os.path.exists(directory):
            shutil.rmtree(directory)
            os.makedirs(directory, exist_ok=True)
            self.logger.info(f"Cleared and reset directory {directory}")



# src/utils/logger.py

import logging
import os
from datetime import datetime
from .singleton import Singleton


class Logger(metaclass=Singleton):
    current_time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

    def __init__(self, name, level, log_directory, log_file):
        self.log_directory = log_directory

        self.logger = logging.getLogger(name)
        self.logger.setLevel(level)

        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(name)s - %(message)s')

        ch = logging.StreamHandler()
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)

        if not os.path.exists(log_directory):
            os.makedirs(log_directory)

        fh = logging.FileHandler(os.path.join(log_directory, log_file))
        fh.setFormatter(formatter)
        self.logger.addHandler(fh)

        self.logger.info(f'Session start at {Logger.current_time}')
        self.logger.info(f'Session start in {os.path.abspath(os.getcwd())}')

    def get_logger(self):
        return self.logger



import torch
from torch.utils.data import DataLoader, Subset, random_split
import torchvision.transforms as transforms
import torchvision.datasets as datasets
import numpy as np
from typing import Optional
import logging
from ..utils import plot_images


class MNISTDataLoader:
    def __init__(self, data_dir, batch_size, shuffle, random_seed, logger):
        self.data_dir: Optional[str] = data_dir
        self.train_loader: Optional[DataLoader] = None
        self.val_loader: Optional[DataLoader] = None
        self.test_loader: Optional[DataLoader] = None

        self.data_name = "MNIST"

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.random_seed = random_seed
        self.logger = logger or logging.getLogger(__name__)
        self.class_names = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']

        self.transform = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,))
        ])

    def display_sample_data(self, data: DataLoader, num_samples) -> None:
        """
        Displays sample images from a DataLoader.
        """
        images, labels = next(iter(data))

        images = images.numpy().transpose((0, 2, 3, 1))
        images = images * 0.3081 + 0.1307

        plot_images(images=images[:num_samples], cls_true=labels[:num_samples], label_names=self.class_names, title="Sample MNIST Images")

        self.logger.info("Sample data displayed successfully.")

    def describe_data(self) -> None:
        """
        Logs a description of the data, detailing the number of samples in each set.
        """
        self.logger.info("Data Description")
        self.logger.info("-----------------")
        if self.train_loader is not None:
            self.logger.info(f"Train Data: Number of samples: {len(self.train_loader.dataset)}")
        if self.val_loader is not None:
            self.logger.info(f"Validation Data: Number of samples: {len(self.val_loader.dataset)}")
        if self.test_loader is not None:
            self.logger.info(f"Test Data: Number of samples: {len(self.test_loader.dataset)}")

    def load_data(self, split_train_val_proportion, train_size, test_size):
        if sum(split_train_val_proportion) != 1.0:
            raise ValueError("Proportions must sum to 1.0")

        np.random.seed(self.random_seed)

        train_dataset = datasets.MNIST(root=self.data_dir, train=True, download=True, transform=self.transform)
        test_dataset = datasets.MNIST(root=self.data_dir, train=False, download=True, transform=self.transform)

        if train_size is None:
            train_size_full = int(split_train_val_proportion[0] * len(train_dataset))
            val_size = len(train_dataset) - train_size_full
        else:
            train_size_full = int(split_train_val_proportion[0] * train_size)
            val_size = train_size - train_size_full

        train_subset, val_subset = random_split(train_dataset, [train_size_full, val_size])

        if train_size is not None and train_size < len(train_subset):
            train_indices = np.random.choice(len(train_subset), train_size_full, replace=False)
            train_subset = Subset(train_subset, train_indices)
            train_subset.dataset.classes = self.class_names

        if test_size is not None and test_size < len(test_dataset):
            test_indices = np.random.choice(len(test_dataset), test_size, replace=False)
            test_dataset = Subset(test_dataset, test_indices)
            test_dataset.dataset.classes = self.class_names

        self.train_loader = DataLoader(train_subset, batch_size=self.batch_size, shuffle=self.shuffle)
        self.val_loader = DataLoader(val_subset, batch_size=self.batch_size, shuffle=False)
        self.test_loader = DataLoader(test_dataset, batch_size=self.batch_size, shuffle=False)

        self.logger.info("Data loaded successfully.")



# src/basic_models/cnn_model.py

import torch
import torch.nn as nn
import torch.nn.functional as F


class SimpleCNN(nn.Module):
    def __init__(self):
        super(SimpleCNN, self).__init__()
        self.conv1 = nn.Conv2d(in_channels=1, out_channels=16, kernel_size=3, padding=1)
        self.conv2 = nn.Conv2d(in_channels=16, out_channels=32, kernel_size=3, padding=1)
        self.fc1 = nn.Linear(in_features=7 * 7 * 32, out_features=128)
        self.fc2 = nn.Linear(in_features=128, out_features=10)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.relu(self.conv1(x))
        x = F.max_pool2d(x, kernel_size=2)
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, kernel_size=2)
        x = x.view(-1, 7 * 7 * 32)
        x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)


class CIFAR10BasicModel(nn.Module):
    def __init__(self):
        super(CIFAR10BasicModel, self).__init__()
        self.conv1 = nn.Conv2d(in_channels=3, out_channels=32, kernel_size=3, padding=1)
        self.conv2 = nn.Conv2d(in_channels=32, out_channels=64, kernel_size=3, padding=1)
        self.relu = nn.ReLU()
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.flatten = nn.Flatten()
        self.fc1 = nn.Linear(in_features=64 * 8 * 8, out_features=512)
        self.fc2 = nn.Linear(in_features=512, out_features=10)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.relu(self.conv1(x))
        x = self.pool(x)
        x = self.relu(self.conv2(x))
        x = self.pool(x)
        x = self.flatten(x)
        x = self.relu(self.fc1(x))
        x = self.fc2(x)
        return x


class ResidualBlock(nn.Module):
    expansion = 1

    def __init__(self, in_channels: int, out_channels: int, stride: int = 1, downsample: nn.Module = None):
        super(ResidualBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(out_channels)
        self.downsample = downsample

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class CustomAlexNet(nn.Module):
    def __init__(self, num_classes: int = 10):
        super(CustomAlexNet, self).__init__()
        self.features = nn.Sequential(
            nn.Conv2d(1, 64, kernel_size=5, stride=1, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),

            nn.Conv2d(64, 192, kernel_size=5, stride=1, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),

            nn.Conv2d(192, 384, kernel_size=3, stride=1, padding=1),
            nn.ReLU(inplace=True),

            nn.Conv2d(384, 256, kernel_size=3, stride=1, padding=1),
            nn.ReLU(inplace=True),

            nn.Conv2d(256, 256, kernel_size=3, stride=1, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),
        )
        self.classifier = nn.Sequential(
            nn.Dropout(),
            nn.Linear(256 * 3 * 3, 1024),
            nn.ReLU(inplace=True),
            nn.Dropout(),
            nn.Linear(1024, 1024),
            nn.ReLU(inplace=True),
            nn.Linear(1024, num_classes),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier(x)
        return x


class CustomResNet(nn.Module):
    def __init__(self, block: nn.Module, layers: list[int], num_classes: int = 10):
        super(CustomResNet, self).__init__()
        self.in_channels = 16
        self.conv1 = nn.Conv2d(1, 16, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(16)
        self.relu = nn.ReLU(inplace=True)
        self.layer1 = self._make_layer(block, 16, layers[0])
        self.layer2 = self._make_layer(block, 32, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 64, layers[2], stride=2)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(64 * block.expansion, num_classes)

    def _make_layer(self, block: nn.Module, out_channels: int, blocks: int, stride: int = 1) -> nn.Sequential:
        downsample = None
        if stride != 1 or self.in_channels != out_channels * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.in_channels, out_channels * block.expansion, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(out_channels * block.expansion),
            )

        layers = []
        layers.append(block(self.in_channels, out_channels, stride, downsample))
        self.in_channels = out_channels * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.in_channels, out_channels))

        return nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)

        return x



import numpy as np


class Individual:
    def __init__(self, net_info, init):
        self.net_info = net_info
        self.gene = np.zeros((self.net_info.node_num + self.net_info.out_num, self.net_info.max_in_num + 1), dtype=int)
        self.is_active = np.empty(self.net_info.node_num + self.net_info.out_num, dtype=bool)
        self.eval = None
        if init:
            print('Initializing with specific architectures...')
            self.init_gene_with_conv()
        else:
            self.init_gene()

    def init_gene_with_conv(self):
        arch = ['S_ConvBlock_64_3']
        input_layer_num = int(self.net_info.input_num / self.net_info.rows) + 1
        output_layer_num = int(self.net_info.out_num / self.net_info.rows) + 1
        layer_ids = [(self.net_info.cols - 1 - input_layer_num - output_layer_num + i) // len(arch) for i in range(len(arch))]
        prev_id, current_layer = 0, input_layer_num
        block_ids = []

        for i, idx in enumerate(layer_ids):
            current_layer += idx
            n = current_layer * self.net_info.rows + np.random.randint(self.net_info.rows)
            block_ids.append(n)
            self.gene[n][0] = self.net_info.func_type.index(arch[i])
            col = min(n // self.net_info.rows, self.net_info.cols)
            max_connect_id = col * self.net_info.rows + self.net_info.input_num
            min_connect_id = max(0, (col - self.net_info.level_back) * self.net_info.rows + self.net_info.input_num)

            self.gene[n][1] = prev_id
            for j in range(1, self.net_info.max_in_num):
                self.gene[n][j + 1] = min_connect_id + np.random.randint(max_connect_id - min_connect_id)

            prev_id = n + self.net_info.input_num

        self.init_output_layer(block_ids, prev_id)
        self.init_intermediate_nodes(block_ids)

    def init_output_layer(self, block_ids, prev_id):
        n = self.net_info.node_num
        type_num = self.net_info.func_type_num if n < self.net_info.node_num else self.net_info.out_type_num
        self.gene[n][0] = np.random.randint(type_num)
        col = min(n // self.net_info.rows, self.net_info.cols)
        max_connect_id = col * self.net_info.rows + self.net_info.input_num
        min_connect_id = max(0, (col - self.net_info.level_back) * self.net_info.rows + self.net_info.input_num)

        self.gene[n][1] = prev_id
        for i in range(1, self.net_info.max_in_num):
            self.gene[n][i + 1] = min_connect_id + np.random.randint(max_connect_id - min_connect_id)

        block_ids.append(n)

    def init_intermediate_nodes(self, block_ids):
        for n in range(self.net_info.node_num + self.net_info.out_num):
            if n in block_ids:
                continue

            type_num = self.net_info.func_type_num if n < self.net_info.node_num else self.net_info.out_type_num
            self.gene[n][0] = np.random.randint(type_num)
            col = min(n // self.net_info.rows, self.net_info.cols)
            max_connect_id = col * self.net_info.rows + self.net_info.input_num
            min_connect_id = max(0, (col - self.net_info.level_back) * self.net_info.rows + self.net_info.input_num)

            for i in range(self.net_info.max_in_num):
                self.gene[n][i + 1] = min_connect_id + np.random.randint(max_connect_id - min_connect_id)

        self.check_active()

    def init_gene(self):
        for n in range(self.net_info.node_num + self.net_info.out_num):
            type_num = self.net_info.func_type_num if n < self.net_info.node_num else self.net_info.out_type_num
            self.gene[n][0] = np.random.randint(type_num)
            col = min(n // self.net_info.rows, self.net_info.cols)
            max_connect_id = col * self.net_info.rows + self.net_info.input_num
            min_connect_id = max(0, (col - self.net_info.level_back) * self.net_info.rows + self.net_info.input_num)

            for i in range(self.net_info.max_in_num):
                self.gene[n][i + 1] = min_connect_id + np.random.randint(max_connect_id - min_connect_id)

        self.check_active()

    def __check_course_to_out(self, n):
        if not self.is_active[n]:
            self.is_active[n] = True
            t = self.gene[n][0]
            in_num = self.net_info.out_in_num[t] if n >= self.net_info.node_num else self.net_info.func_in_num[t]

            for i in range(in_num):
                if self.gene[n][i + 1] >= self.net_info.input_num:
                    self.__check_course_to_out(self.gene[n][i + 1] - self.net_info.input_num)

    def check_active(self):
        self.is_active[:] = False
        for n in range(self.net_info.out_num):
            self.__check_course_to_out(self.net_info.node_num + n)

    def check_pool(self):
        pool_num = sum(1 for n in range(self.net_info.node_num + self.net_info.out_num) if self.is_active[n] and self.gene[n][0] > 19)
        return pool_num == 0, pool_num

    def __mutate(self, current, min_int, max_int):
        mutated_gene = current
        while current == mutated_gene:
            mutated_gene = min_int + np.random.randint(max_int - min_int)
        return mutated_gene

    def mutation(self, mutation_rate=0.01):
        active_check = False
        for n in range(self.net_info.node_num + self.net_info.out_num):
            t = self.gene[n][0]
            type_num = self.net_info.func_type_num if n < self.net_info.node_num else self.net_info.out_type_num
            if np.random.rand() < mutation_rate and type_num > 1:
                self.gene[n][0] = self.__mutate(self.gene[n][0], 0, type_num)
                if self.is_active[n]:
                    active_check = True

            col = min(n // self.net_info.rows, self.net_info.cols)
            max_connect_id = col * self.net_info.rows + self.net_info.input_num
            min_connect_id = max(0, (col - self.net_info.level_back) * self.net_info.rows + self.net_info.input_num)
            in_num = self.net_info.func_in_num[t] if n < self.net_info.node_num else self.net_info.out_in_num[t]

            for i in range(self.net_info.max_in_num):
                if np.random.rand() < mutation_rate and max_connect_id - min_connect_id > 1:
                    self.gene[n][i + 1] = self.__mutate(self.gene[n][i + 1], min_connect_id, max_connect_id)
                    if self.is_active[n] and i < in_num:
                        active_check = True

        self.check_active()
        return active_check

    def neutral_mutation(self, mutation_rate=0.01):
        for n in range(self.net_info.node_num + self.net_info.out_num):
            t = self.gene[n][0]
            type_num = self.net_info.func_type_num if n < self.net_info.node_num else self.net_info.out_type_num
            if not self.is_active[n] and np.random.rand() < mutation_rate and type_num > 1:
                self.gene[n][0] = self.__mutate(self.gene[n][0], 0, type_num)

            col = min(n // self.net_info.rows, self.net_info.cols)
            max_connect_id = col * self.net_info.rows + self.net_info.input_num
            min_connect_id = max(0, (col - self.net_info.level_back) * self.net_info.rows + self.net_info.input_num)
            in_num = self.net_info.func_in_num[t] if n < self.net_info.node_num else self.net_info.out_in_num[t]

            for i in range(self.net_info.max_in_num):
                if (not self.is_active[n] or i >= in_num) and np.random.rand() < mutation_rate and max_connect_id - min_connect_id > 1:
                    self.gene[n][i + 1] = self.__mutate(self.gene[n][i + 1], min_connect_id, max_connect_id)

        self.check_active()

    def count_active_node(self):
        return self.is_active.sum()

    def copy(self, source):
        self.net_info = source.net_info
        self.gene = source.gene.copy()
        self.is_active = source.is_active.copy()
        self.eval = source.eval

    def active_net_list(self):
        net_list = [["input", 0, 0]]
        active_cnt = np.arange(self.net_info.input_num + self.net_info.node_num + self.net_info.out_num)
        active_cnt[self.net_info.input_num:] = np.cumsum(self.is_active)

        for n, is_a in enumerate(self.is_active):
            if is_a:
                t = self.gene[n][0]
                type_str = self.net_info.func_type[t] if n < self.net_info.node_num else self.net_info.out_type[t]
                connections = [active_cnt[self.gene[n][i + 1]] for i in range(self.net_info.max_in_num)]
                net_list.append([type_str] + connections)
        return net_list



import torch
from torch.utils.data import DataLoader
import torchvision.transforms as transforms
from ..utils import plot_images
import torchvision.datasets as datasets
import logging
import os
import numpy as np
from typing import Optional


class ImageNetDataLoader:
    def __init__(self, data_dir, scraper, batch_size, shuffle, random_seed, image_size, logger):
        self.data_dir: Optional[str] = data_dir
        self.train_loader: Optional[DataLoader] = None
        self.val_loader: Optional[DataLoader] = None
        self.test_loader: Optional[DataLoader] = None

        self.data_name = "ImageNet"

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.random_seed = random_seed
        self.logger = logger or logging.getLogger(__name__)

        self.transform = transforms.Compose([
            transforms.Resize(image_size),
            transforms.ToTensor(),
            transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225))
        ])

        self.scraper = scraper

    def display_sample_data(self, data: DataLoader, num_samples: int) -> None:
        """
        Displays sample images from a DataLoader.
        """
        images, labels = next(iter(data))

        images = images.numpy().transpose((0, 2, 3, 1))
        images = images * np.array([0.229, 0.224, 0.225]) + np.array([0.485, 0.456, 0.406])

        plot_images(images=images[:num_samples], cls_true=labels[:num_samples], label_names=data.dataset.classes, title="Sample ImageNet Images")

        self.logger.info("Sample data displayed successfully.")

    def describe_data(self) -> None:
        """
        Logs a description of the data, detailing the number of samples in each set.
        """
        self.logger.info("Data Description")
        self.logger.info("-----------------")
        if self.train_loader is not None:
            self.logger.info(f"Train Data: Number of samples: {len(self.train_loader.dataset)}")
        if self.val_loader is not None:
            self.logger.info(f"Validation Data: Number of samples: {len(self.val_loader.dataset)}")
        if self.test_loader is not None:
            self.logger.info(f"Test Data: Number of samples: {len(self.test_loader.dataset)}")

    def load_data(self):
        """
        Downloads and splits the data using ImageNetScraper, then loads it into DataLoader objects.
        """
        train_dataset = datasets.ImageFolder(os.path.join(self.data_dir, 'train'), transform=self.transform)
        val_dataset = datasets.ImageFolder(os.path.join(self.data_dir, 'valid'), transform=self.transform)
        test_dataset = datasets.ImageFolder(os.path.join(self.data_dir, 'test'), transform=self.transform)

        self.train_loader = DataLoader(train_dataset, batch_size=self.batch_size, shuffle=self.shuffle)
        self.val_loader = DataLoader(val_dataset, batch_size=self.batch_size, shuffle=False)
        self.test_loader = DataLoader(test_dataset, batch_size=self.batch_size, shuffle=False)

        self.logger.info("Data loaded successfully.")

        return self.train_loader, self.val_loader, self.test_loader



import numpy as np


class CgpConfig:
    def __init__(self, rows, cols, level_back, min_active_num, max_active_num, func_type, func_in_num):
        self.input_num = 1

        self.func_type = func_type
        self.func_in_num = func_in_num

        self.out_num = 1
        self.out_type = ['full']
        self.out_in_num = [1]

        self.rows = rows
        self.cols = cols
        self.node_num = rows * cols
        self.level_back = level_back
        self.min_active_num = min_active_num
        self.max_active_num = max_active_num

        self.func_type_num = len(self.func_type)
        self.out_type_num = len(self.out_type)
        self.max_in_num = max(np.max(self.func_in_num), np.max(self.out_in_num))



import time
import torch
import torch.optim as optim
from enum import Enum
import os
import mlflow
import mlflow.pytorch
import seaborn as sns
from sklearn.metrics import confusion_matrix
import matplotlib.pyplot as plt


class OptimizerType(Enum):
    ADAM = 'Adam'
    SGD = 'SGD'
    RMSPROP = 'RMSprop'


class CNNTrainer:
    def __init__(self, train_loader, test_loader, val_loader, verbose, mlflow_enabled, experiment_name, logger, channel, imgSize, batchsize, result_folder_path, ml_flow_folder_path):
        self.verbose = verbose
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.train_loader = train_loader
        self.val_loader = val_loader
        self.test_loader = test_loader
        self.mlflow_enabled = mlflow_enabled
        self.experiment_name = experiment_name
        self.logger = logger
        self.channel = channel
        self.imgSize = imgSize
        self.batchsize = batchsize
        self.result_folder_path = result_folder_path
        self.ml_flow_folder_path = ml_flow_folder_path

        if self.mlflow_enabled:
            self.init_mlflow()

    def train(self, model, learning_rate, optimizer_type, criterion, early_stopping, patience, epochs):
        model = model.to(self.device)
        model.train()

        best_loss = float('inf')
        patience_counter = 0

        criterion = criterion.to(self.device)
        optimizer = self.configure_optimizer(model, learning_rate, optimizer_type)

        input = torch.FloatTensor(self.batchsize, self.channel, self.imgSize, self.imgSize).to(self.device)
        label = torch.LongTensor(self.batchsize).to(self.device)

        history = []

        if self.verbose:
            self.logger.info(f"Training for {epochs} epochs.")

        start_train = time.time()

        for epoch in range(epochs):
            running_loss = 0.0

            for batch_idx, (data, target) in enumerate(self.train_loader):
                data, target = data.to(self.device), target.to(self.device)

                input.resize_as_(data).copy_(data)
                label.resize_as_(target).copy_(target)

                optimizer.zero_grad()
                output = model(data)
                loss = criterion(output, target)
                loss.backward()
                optimizer.step()

                running_loss += loss.item()

            avg_loss = running_loss / len(self.train_loader)
            history.append(avg_loss)

            if early_stopping:
                if avg_loss < best_loss:
                    best_loss = avg_loss
                    patience_counter = 0
                else:
                    patience_counter += 1
                    if patience_counter >= patience:
                        if self.verbose:
                            self.logger.info(f"Early stopping triggered at epoch {epoch + 1}")
                        break

        end_train = time.time()

        if self.verbose:
            self.logger.info(f"Training completed in {end_train - start_train:.2f} seconds.")

        return model, history

    def evaluate(self, model):
        correct = 0
        total = 0
        model.eval()

        start_eval = time.time()

        with torch.no_grad():
            for data, target in self.train_loader:
                data, target = data.to(self.device), target.to(self.device)
                output = model(data)
                _, predicted = torch.max(output.data, 1)
                total += target.size(0)
                correct += (predicted == target).sum().item()

        end_eval = time.time()
        accuracy = correct / total

        if self.verbose:
            self.logger.info(f"Evaluation completed in {end_eval - start_eval:.2f} seconds. Accuracy: {accuracy * 100:.2f}%")

        return accuracy

    def predict(self, model, images):
        model = model.to(self.device)
        model.eval()

        images = images.to(self.device)

        with torch.no_grad():
            outputs = model(images)
            _, predicted = torch.max(outputs.data, 1)

        return predicted.cpu().numpy()

    def __call__(self, model, learning_rate, optimizer_type, criterion, early_stopping, patience, epochs):
        trained_model, history = self.train(
            model=model,
            learning_rate=learning_rate,
            optimizer_type=optimizer_type,
            criterion=criterion,
            early_stopping=early_stopping,
            patience=patience,
            epochs=epochs
        )
        accuracy = self.evaluate(trained_model)

        return trained_model, accuracy, history

    def save_model_and_results(self, model, name):
        os.makedirs(self.result_folder_path, exist_ok=True)

        model_path = os.path.join(self.result_folder_path, f'{name}.pth')
        torch.save(model.state_dict(), model_path)
        self.logger.info(f"Model saved to {model_path}")

        if self.mlflow_enabled:
            mlflow.pytorch.log_model(model, self.ml_flow_folder_path)

        model.to(self.device)
        model.eval()

        all_preds = []
        all_targets = []
        correct = 0
        total = 0

        with torch.no_grad():
            for data, target in self.test_loader:
                data, target = data.to(self.device), target.to(self.device)
                output = model(data)
                _, predicted = torch.max(output.data, 1)

                all_preds.extend(predicted.cpu().numpy())
                all_targets.extend(target.cpu().numpy())

                total += target.size(0)
                correct += (predicted == target).sum().item()

        accuracy = correct / total
        self.logger.info(f"Model accuracy: {accuracy * 100:.2f}%")

        cm = confusion_matrix(all_targets, all_preds)
        plt.figure(figsize=(10, 7))
        sns.heatmap(cm, annot=True, fmt="d", cmap="Blues", xticklabels=self.test_loader.dataset.classes, yticklabels=self.test_loader.dataset.classes)
        title = f'Confusion Matrix for {name}\nAccuracy: {accuracy * 100:.2f}%'
        plt.title(title)
        plt.ylabel('True label')
        plt.xlabel('Predicted label')
        plt.savefig(os.path.join(self.result_folder_path, 'confusion_matrix.png'))
        self.logger.info(f"Confusion matrix saved to {os.path.join(self.result_folder_path, 'confusion_matrix.png')}")
        plt.close()

    def init_mlflow(self) -> None:
        encoded_uri = f"file:///{self.ml_flow_folder_path}"
        mlflow.set_tracking_uri(encoded_uri)
        mlflow.set_experiment(self.experiment_name)
        self.logger.info(f"MLflow URI: {encoded_uri}")
        self.logger.info(f"MLflow configured with tracking URI: {self.ml_flow_folder_path} and experiment name: {self.experiment_name}")

    def configure_optimizer(self, model, learning_rate, optimizer_type):
        if optimizer_type == OptimizerType.ADAM:
            return optim.Adam(model.parameters(), lr=learning_rate, betas=(0.5, 0.999))
        elif optimizer_type == OptimizerType.SGD:
            return optim.SGD(model.parameters(), lr=learning_rate, momentum=0.9)
        elif optimizer_type == OptimizerType.RMSPROP:
            return optim.RMSprop(model.parameters(), lr=learning_rate)
        else:
            raise ValueError(f"Unsupported optimizer: {optimizer_type}")



import torch
import torch.nn as nn
from torch.autograd import Variable
import torch.nn.functional as F
import math
from enum import Enum


class DropoutRate(Enum):
    NONE = 0
    RATE_25 = 0.25
    RATE_50 = 0.5
    RATE_75 = 0.75


class ConvBlock(nn.Module):
    def __init__(self, in_size, out_size, kernel, stride, dropout_rate=DropoutRate.NONE):
        super(ConvBlock, self).__init__()
        pad_size = kernel // 2
        self.conv1 = nn.Sequential(
            nn.Conv2d(in_size, out_size, kernel, stride=stride, padding=pad_size, bias=False),
            nn.BatchNorm2d(out_size),
            nn.ReLU(inplace=True),
            nn.Dropout(dropout_rate.value)
        )

    def forward(self, inputs):
        return self.conv1(inputs)


class ConvBlock_sum(nn.Module):
    def __init__(self, in_size, out_size, kernel, dropout_rate=DropoutRate.NONE):
        super(ConvBlock_sum, self).__init__()
        pad_size = kernel // 2
        self.conv1 = nn.Sequential(
            nn.Conv2d(in_size, out_size, kernel, padding=pad_size, bias=False),
            nn.BatchNorm2d(out_size),
            nn.ReLU(inplace=True),
            nn.Dropout(dropout_rate.value)
        )
        self.relu = nn.ReLU(inplace=True)

    def forward(self, inputs1, inputs2):
        outputs = self.conv1(inputs1)
        in_data = [outputs, inputs2]
        small_ch_id, large_ch_id = (0, 1) if in_data[0].size(1) < in_data[1].size(1) else (1, 0)
        offset = int(in_data[large_ch_id].size()[1] - in_data[small_ch_id].size()[1])
        if offset != 0:
            tmp = in_data[large_ch_id].data[:, :offset, :, :]
            tmp = Variable(tmp).clone()
            in_data[small_ch_id] = torch.cat([in_data[small_ch_id], tmp * 0], 1)
        return self.relu(torch.add(in_data[0], in_data[1]))


class ResBlock(nn.Module):
    def __init__(self, in_size, out_size, kernel, stride, dropout_rate=DropoutRate.NONE):
        super(ResBlock, self).__init__()
        pad_size = kernel // 2
        self.conv1 = nn.Sequential(
            nn.Conv2d(in_size, out_size, kernel, stride=stride, padding=pad_size, bias=False),
            nn.BatchNorm2d(out_size),
            nn.ReLU(inplace=True),
            nn.Dropout(dropout_rate.value),
            nn.Conv2d(out_size, out_size, kernel, stride=stride, padding=pad_size, bias=False),
            nn.BatchNorm2d(out_size),
            nn.Dropout(dropout_rate.value)
        )
        self.relu = nn.ReLU(inplace=True)

    def forward(self, inputs1, inputs2):
        x = self.conv1(inputs1)
        in_data = [x, inputs2]
        small_ch_id, large_ch_id = (0, 1) if in_data[0].size(1) < in_data[1].size(1) else (1, 0)
        offset = int(in_data[large_ch_id].size()[1] - in_data[small_ch_id].size()[1])
        if offset != 0:
            tmp = in_data[large_ch_id].data[:, :offset, :, :]
            tmp = Variable(tmp).clone()
            in_data[small_ch_id] = torch.cat([in_data[small_ch_id], tmp * 0], 1)
        return self.relu(torch.add(in_data[0], in_data[1]))


class Sum(nn.Module):
    def __init__(self):
        super(Sum, self).__init__()

    def forward(self, inputs1, inputs2):
        in_data = [inputs1, inputs2]
        small_in_id, large_in_id = (0, 1) if in_data[0].size(2) < in_data[1].size(2) else (1, 0)
        pool_num = math.floor(in_data[large_in_id].size(2) / in_data[small_in_id].size(2))
        for _ in range(pool_num - 1):
            in_data[large_in_id] = F.max_pool2d(in_data[large_in_id], 2, 2, 0)
        small_ch_id, large_ch_id = (0, 1) if in_data[0].size(1) < in_data[1].size(1) else (1, 0)
        offset = int(in_data[large_ch_id].size()[1] - in_data[small_ch_id].size()[1])
        if offset != 0:
            tmp = in_data[large_ch_id].data[:, :offset, :, :]
            tmp = Variable(tmp).clone()
            in_data[small_ch_id] = torch.cat([in_data[small_ch_id], tmp * 0], 1)
        return torch.add(in_data[0], in_data[1])


class Concat(nn.Module):
    def __init__(self):
        super(Concat, self).__init__()

    def forward(self, inputs1, inputs2):
        in_data = [inputs1, inputs2]
        small_in_id, large_in_id = (0, 1) if in_data[0].size(2) < in_data[1].size(2) else (1, 0)
        pool_num = math.floor(in_data[large_in_id].size(2) / in_data[small_in_id].size(2))
        for _ in range(pool_num - 1):
            in_data[large_in_id] = F.max_pool2d(in_data[large_in_id], 2, 2, 0)
        return torch.cat([in_data[0], in_data[1]], 1)


class LinearBlock(nn.Module):
    def __init__(self, in_features, out_features):
        super(LinearBlock, self).__init__()
        self.linear = nn.Linear(in_features, out_features)

    def forward(self, x):
        return self.linear(x)


class MaxPoolingBlock(nn.Module):
    def __init__(self, kernel_size, stride=None, padding=0):
        super(MaxPoolingBlock, self).__init__()
        self.pool = nn.MaxPool2d(kernel_size, stride=stride, padding=padding)

    def forward(self, x):
        return self.pool(x)


class AvgPoolingBlock(nn.Module):
    def __init__(self, kernel_size, stride=None, padding=0):
        super(AvgPoolingBlock, self).__init__()
        self.pool = nn.AvgPool2d(kernel_size, stride=stride, padding=padding)

    def forward(self, x):
        return self.pool(x)



from .cgp_config import CgpConfig
from .cnn_evolution import CNNEvaluation
from .individual import Individual
from .cgp import CGP

__all__ = [
    'CgpConfig',
    'CNNEvaluation',
    'Individual',
    'CGP'
]



from setuptools import find_packages, setup
from collections import defaultdict


def _cut_version_number_from_requirement(req):
    return req.split()[0]


def read_requirements():
    requirements = []
    with open("./requirements.txt") as f:
        for req in f:
            req = req.replace("\n", " ").strip()
            if req and not req.startswith("#"):  # Pomija puste linie i komentarze
                requirements.append(req)
    return requirements


def read_extra_requirements():
    extra_requirements = defaultdict(lambda: [])
    collect_key = None
    collect_keys = []
    with open("./extra-requirements.txt") as f:
        for req in f:
            req = req.replace("\n", " ").strip()
            if req.startswith("#"):
                collect_key = req.split(" ")[1]
                collect_keys.append(collect_key)
                continue
            if collect_key and req:
                extra_requirements[collect_key].append(req)
    for key in collect_keys:
        extra_requirements["all"] += extra_requirements[key]
    return extra_requirements


with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="cgpcnn",
    version="0.1.1",
    description="A project featuring methods for optimizing neural networks using PyTorch.",
    package_dir={"cgpcnn": "src"},
    packages=["cgpcnn." + pkg for pkg in find_packages("src")],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MrLipa/CGP-CNN-Optimizer",
    author="Tomasz Szkaradek",
    author_email="bilbo.weirdo@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    install_requires=read_requirements(),
    extras_require=read_extra_requirements(),
    python_requires=">=3.10.14",
)



# src/loaders/__init__.py

from .cifar10_loader import CIFAR10DataLoader
from .mnist_loader import MNISTDataLoader
from .image_net_loader import ImageNetDataLoader

__all__ = ['CIFAR10DataLoader', 'MNISTDataLoader', 'ImageNetDataLoader']


import matplotlib.pyplot as plt
import pydot
import matplotlib.animation as animation
from PIL import Image
import io


def plot_images(images, cls_true, cls_pred=None, title=None, label_names=None):
    fig, axes = plt.subplots(3, 3, figsize=(10, 10))
    if title:
        fig.suptitle(title, size=20)

    for i, ax in enumerate(axes.flat):
        ax.imshow(images[i, :, :, :], interpolation='spline16')
        cls_true_name = label_names[int(cls_true[i])]
        xlabel = f"{cls_true_name} ({int(cls_true[i])})"
        if cls_pred is not None:
            cls_pred_name = label_names[int(cls_pred[i])]
            xlabel += f"\nPred: {cls_pred_name}"
        ax.set_xlabel(xlabel)
        ax.set_xticks([])
        ax.set_yticks([])

    plt.show()


def plot_graph(individual, rankdir='TB'):
    graph = pydot.Dot(graph_type='digraph', rankdir=rankdir)
    nodes = {}

    for i in range(individual.net_info.input_num):
        node = pydot.Node(f'input-{i}', label=f'Input {i}', shape='ellipse')
        nodes[i] = node
        graph.add_node(node)

    for idx in range(individual.net_info.node_num + individual.net_info.out_num):
        if individual.is_active[idx]:
            gene_info = individual.gene[idx]
            func_type = individual.net_info.func_type[gene_info[0]] if idx < individual.net_info.node_num else f'Output {idx - individual.net_info.node_num}'
            label = f"{func_type} (Node {idx})"
            node = pydot.Node(f'node-{idx}', label=label, shape='box')
            nodes[idx + individual.net_info.input_num] = node
            graph.add_node(node)

    for idx in range(individual.net_info.node_num + individual.net_info.out_num):
        if individual.is_active[idx]:
            for j in range(1, individual.net_info.max_in_num + 1):
                input_idx = individual.gene[idx][j]
                if input_idx >= 0:
                    edge = pydot.Edge(nodes[input_idx], nodes[idx + individual.net_info.input_num])
                    graph.add_edge(edge)

    png_str = graph.create(prog='dot', format='png')
    image = Image.open(io.BytesIO(png_str))

    return image


def plot_cartesian(individual):
    graph = pydot.Dot(graph_type='digraph')
    nodes = {}

    for i in range(individual.net_info.input_num):
        node = pydot.Node(f'input-{i}', shape='ellipse')
        node.set('pos', f'-1,{i:.2f}!')
        nodes[i] = node
        graph.add_node(node)

    for idx in range(individual.net_info.node_num):
        x = (idx // individual.net_info.rows)
        y = idx % individual.net_info.rows

        node_style = "filled" if individual.is_active[idx] else "filled"
        node_color = '#ff00cc' if individual.is_active[idx] else '#cccccc'
        node = pydot.Node(f'{idx}', label=f'{idx}', style=node_style, shape="circle", fillcolor=node_color)
        node.set('pos', f'{x},{y}!')
        node.set('fontsize', 10)
        nodes[idx + individual.net_info.input_num] = node
        graph.add_node(node)

    for idx in range(individual.net_info.out_num):
        out_idx = individual.net_info.node_num + idx
        node = pydot.Node(f'output-{idx}', label=f'output-{idx}', style="filled", shape="ellipse", fillcolor='#ccaadd')
        node.set('pos', f'{(individual.net_info.cols + 1)},{idx:.2f}!')
        nodes[out_idx + individual.net_info.input_num] = node
        graph.add_node(node)

        for con in range(individual.net_info.max_in_num):
            input_idx = individual.gene[out_idx][con + 1]
            if input_idx in nodes:
                graph.add_edge(pydot.Edge(nodes[input_idx], nodes[out_idx + individual.net_info.input_num]))

    for idx in range(individual.net_info.node_num):
        if individual.is_active[idx]:
            for con in range(individual.net_info.max_in_num):
                input_idx = individual.gene[idx][con + 1]
                if input_idx in nodes:
                    graph.add_edge(pydot.Edge(nodes[input_idx], nodes[idx + individual.net_info.input_num]))

    png_str = graph.create(prog='neato', format='png')
    image = Image.open(io.BytesIO(png_str))

    return image


def plot_combined(individual):
    graph_image = plot_graph(individual)
    cartesian_image = plot_cartesian(individual)

    fig, axs = plt.subplots(2, 1, figsize=(18, 12))

    axs[0].imshow(graph_image)
    axs[0].axis('off')
    axs[0].set_title("Graph Representation", fontsize=16)

    axs[1].imshow(cartesian_image)
    axs[1].axis('off')
    axs[1].set_title("Cartesian Grid Representation", fontsize=16)
    plt.tight_layout()

    fig.canvas.draw()
    combined_image = Image.frombytes('RGBA', fig.canvas.get_width_height(), fig.canvas.buffer_rgba())

    plt.close(fig)

    return combined_image


class ImageAnimation:
    def __init__(self, image_list):
        self.image_list = image_list
        self.iterations = len(image_list)

        self.fig, self.ax = plt.subplots()
        self.im = self.ax.imshow(self.image_list[0], animated=True)

        self.ax.axis('off')
        self.anim = animation.FuncAnimation(
            self.fig, self._draw_frame, frames=self.iterations,
            init_func=self._init, interval=1000, blit=True)

    def _init(self):
        self.im.set_data(self.image_list[0])
        return self.im,

    def _draw_frame(self, framedata):
        self.im.set_data(self.image_list[framedata])
        return self.im,

    def save(self, filename):
        self.anim.save(filename, writer='pillow')



import numpy as np
from torch import nn
from ..trainers import OptimizerType
from ..converters import CGP2CNN


class CNNEvaluation:
    def __init__(self, trainer, logger, verbose=True, epoch_num=25, batchsize=16, imgSize=32):
        self.epoch_num = epoch_num
        self.logger = logger
        self.batchsize = batchsize
        self.verbose = verbose
        self.imgSize = imgSize
        self.trainer = trainer

    def __call__(self, net_lists):
        evaluations = np.zeros(len(net_lists))
        for i in range(len(net_lists)):
            model = CGP2CNN(net_lists[i], 3, 10, 32)
            trained_model, accuracy, history = self.trainer(model, 0.01, OptimizerType.ADAM, nn.CrossEntropyLoss(), False, 5, self.epoch_num)
            evaluations[i] = accuracy
        return evaluations



# src/utils/__init__.py

from .logger import Logger
from .singleton import Singleton
from .image_net_scraper import ImageNetScraper
from .plotting import plot_images, plot_graph, plot_cartesian, plot_combined, ImageAnimation

__all__ = [
    'Logger',
    'Singleton',
    'ImageNetScraper',
    'plot_images',
    'plot_graph',
    'plot_cartesian',
    'plot_combined',
    'ImageAnimation'
]


import torch.nn as nn


def weights_init(m):
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        m.weight.data.normal_(0.0, 0.02)
    elif classname.find('BatchNorm') != -1:
        m.weight.data.normal_(1.0, 0.02)
        m.bias.data.fill_(0)


def weights_init_normal(m):
    classname = m.__class__.__name__
    if classname.find('Conv2d') != -1:
        m.apply(weights_init_normal_)
    elif classname.find('Linear') != -1:
        nn.init.uniform_(m.weight.data, 0.0, 0.02)
    elif classname.find('BatchNorm2d') != -1:
        nn.init.uniform_(m.weight.data, 1.0, 0.02)
        nn.init.constant_(m.bias.data, 0.0)


def weights_init_normal_(m):
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        nn.init.uniform_(m.weight.data, 0.0, 0.02)
    elif classname.find('Linear') != -1:
        nn.init.uniform_(m.weight.data, 0.0, 0.02)
    elif classname.find('BatchNorm2d') != -1:
        nn.init.uniform_(m.weight.data, 1.0, 0.02)
        nn.init.constant_(m.bias.data, 0.0)


def weights_init_xavier(m):
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        nn.init.xavier_normal_(m.weight.data, gain=1)
    elif classname.find('Linear') != -1:
        nn.init.xavier_normal_(m.weight.data, gain=1)
    elif classname.find('BatchNorm2d') != -1:
        nn.init.uniform_(m.weight.data, 1.0, 0.02)
        nn.init.constant_(m.bias.data, 0.0)


def weights_init_kaiming(m):
    classname = m.__class__.__name__
    if classname.find('Conv2d') != -1:
        nn.init.kaiming_normal_(m.weight.data, a=0, mode='fan_in')
    elif classname.find('Linear') != -1:
        nn.init.kaiming_normal_(m.weight.data, a=0, mode='fan_in')
    elif classname.find('BatchNorm2d') != -1:
        nn.init.uniform_(m.weight.data, 0.02, 1.0)
        nn.init.constant_(m.bias.data, 0.0)


def weights_init_orthogonal(m):
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        nn.init.orthogonal_(m.weight.data, gain=1)
    elif classname.find('Linear') != -1:
        nn.init.orthogonal_(m.weight.data, gain=1)
    elif classname.find('BatchNorm2d') != -1:
        nn.init.uniform_(m.weight.data, 1.0, 0.02)
        nn.init.constant_(m.bias.data, 0.0)


def init_weights(net, init_type='normal'):
    if init_type == 'normal':
        net.apply(weights_init_normal)
    elif init_type == 'xavier':
        net.apply(weights_init_xavier)
    elif init_type == 'kaiming':
        net.apply(weights_init_kaiming)
    elif init_type == 'orthogonal':
        net.apply(weights_init_orthogonal)
    else:
        raise NotImplementedError(f'Initialization method [{init_type}] is not implemented')
